#include<stdio.h>
#include<stdlib.h>
#include<stdbool.h>
#include<string.h>
#include<limits.h>
#include<time.h>

#define UNUSED(x) (void)x;
#define PRINT_DEBUG 0

#define BENCHING 1

// 0 off, 1 cost only, 2 cost and combination
#define PRINT_COMBINATION 1

const int MaxWeight = 8*60;

typedef struct Node
{
	char* Label;
	int Weight;
} Node;

typedef struct ListItem ListItem;

struct ListItem {
	ListItem* Previous;
	ListItem* Next;
	ListItem* NextAtDepth;
	int Weight;
	int Index;
	int Depth;
};
ListItem* InsertAfterp(ListItem* item, ListItem* newItem)
{

	ListItem* last = item->Next;
	newItem->Previous = item;
	newItem->Next = last;
	item->Next = newItem;
	if (last != NULL)
		last->Previous = newItem;
	return newItem;
}
ListItem* InsertAfter(ListItem* item, int weight, int index, int depth)
{
	ListItem* newItem = malloc(sizeof(ListItem));
	newItem->Index = index;
	newItem->Weight = weight;
	newItem->Depth = depth;
	newItem->NextAtDepth = NULL;
	return InsertAfterp(item, newItem);
}

ListItem* InsertSorted(ListItem* start, int weight, int index, int depth)
{
	ListItem* it = start;
	while (it->Next != NULL && it->Next->Weight > weight)
	{
		it = it->Next;
	}
	return InsertAfter(it, weight, index, depth);
}
ListItem* InsertSortedp(ListItem* start, ListItem* item)
{
	ListItem* it = start;
	int weight = item->Weight;
	while (it->Next != NULL && it->Next->Weight > weight)
	{
		it = it->Next;
	}
	return InsertAfterp(it, item);
}

void RemoveItem(ListItem* item)
{
	ListItem* before = item->Previous;
	ListItem* after = item->Next;
	if (before != NULL)
		before->Next = after;
	if (after != NULL)
		after->Previous = before;
	item->Next = NULL;
	item->Previous = NULL;
	free(item);
}
void DetachItem(ListItem* item)
{
	ListItem* before = item->Previous;
	ListItem* after = item->Next;
	if (before != NULL)
		before->Next = after;
	if (after != NULL)
		after->Previous = before;
}
void AttachItem(ListItem* item)
{
	ListItem* before = item->Previous;
	ListItem* after = item->Next;
	if (before != NULL)
		before->Next = item;
	if (after != NULL)
		after->Previous = item;
}


void DeinitNode(Node* node)
{
	if (node->Label != NULL)
		free(node->Label);
}

typedef struct Graph
{
	Node* Nodes;
	int Count;

	int* EdgeCounters;
	int* EdgeCounts;
	int** Edges;
} Graph;

Graph* InitGraph(int v)
{
	Graph* g = malloc(sizeof(Graph));
	g->Nodes = malloc(sizeof(Node)*v);
	g->Count = v;
	g->Edges = malloc(sizeof(int*)*v);
	g->EdgeCounters = malloc(sizeof(int)*v);
	g->EdgeCounts = malloc(sizeof(int)*v);
	for (int i = 0; i < v; i++)
	{
		g->Edges[i] = NULL;
		g->Nodes[i].Label = NULL;
		g->EdgeCounters[i] = 0;
		
	}
	return g;
}
void GenerateEdgeData(Graph* g)
{
	int c = g->Count;
	
	for (int i = 0; i < c;i++)
	{
		for (int j = 0; j < g->EdgeCounts[i]; j++)
		{
			g->EdgeCounters[g->Edges[i][j]]++;
		}
	}
}
void DeinitGraph(Graph* graph)
{
	int v = graph->Count;
	for (int i = 0; i < v; i++)
	{
		DeinitNode(&graph->Nodes[i]);
		if (graph->Edges[i] != NULL)
			free(graph->Edges[i]);
	}
	free(graph->Edges);
	free(graph->EdgeCounters);
	free(graph->EdgeCounts);
	free(graph->Nodes);
	free(graph);
}

void BranchAndCut(Graph* graph);


int random(int lower, int upper)
{
	// This technique is instead of the usual modulus, to maintain uniformity.
	float r = ((float)rand())/((float)RAND_MAX);
	int diff = upper-lower;
	int r_int = (int)(r*(float)diff);
	return r_int+lower;
}

void uniqueConstraints(int* array, int maxValue, int count)
{
	int vals[maxValue];
	for (int i = 0; i < maxValue; i++)
		vals[i] = i;

	for (int i = 0; i < maxValue;i++)
	{
		int idx = random(i,maxValue);
		int tmp = vals[i];
		vals[i] = vals[idx];
		vals[idx] = tmp;
	}
	for (int i = 0; i < count;i++)
	{
		array[i] = vals[i];
	}
}


Graph* GenerateRandomizedProblem(int seed, int items, int constraints, float sizeLower, float sizeUpper)
{
	int sizeLow = (int)(sizeLower*(float)MaxWeight);
	int sizeUp = (int)(sizeUpper*(float)MaxWeight);
	srand(seed);
	
	Graph* g = InitGraph(items);
	for (int i = 0; i < items; i++)
	{
		int length = random(sizeLow, sizeUp);
		g->Nodes[i].Weight = length;
		int cons = (constraints*i)/100;
		g->Edges[i] = malloc(sizeof(int)*cons);
		g->EdgeCounts[i] = cons;
		if (cons > 0)
			uniqueConstraints(g->Edges[i], i, cons);
	}
	GenerateEdgeData(g);
	return g;

}

Graph* GenerateDepthStudyProblemInstance(int scenes, int reps)
{
	Graph* g = InitGraph(scenes*reps);
	for (int i = 0; i < scenes; i++)
	{
		int length = ((i % 7)+2)*15;
		for (int j = 0; j < reps;j++)
		{
			int idx = i*reps+j;
			g->Nodes[idx].Weight = length;
			

			if (j != reps-1)
			{
				g->Edges[idx] = malloc(sizeof(int)*1);
				g->EdgeCounts[idx] = 1;
				g->Edges[idx][0] = idx+1;
			}
			else
			{
				g->EdgeCounts[idx] = 0;
				g->Edges[idx] = malloc(sizeof(int)*0);
			}
		}
	}
	GenerateEdgeData(g);
	
#if PRINT_DEBUG
	int c = g->Count;
	for (int i = 0; i < c; i++)
	{
		printf("Edges from node %i:", i);
		for (int j = 0; j < g->EdgeCounts[i]; j++)
		{
			printf(" %i", g->Edges[i][j]);
		}
		printf("\n");
	}

	for (int i = 0; i < c; i++)
	{
		printf("EdgeCounter[%i]: %i\n", i, g->EdgeCounters[i]);
	}
#endif
	return g;
}

clock_t start;
int timePassed(void)
{
	clock_t stop = clock();
	return (int)(stop-start)/(CLOCKS_PER_SEC/1000);
}

int main(int argv, char** argc)
{
	#if BENCHING//print immediatly, to avoid lack of output when terminating with timeout-tool.
	setbuf(stdout, NULL);
	#endif
	if (argv < 2)
	{
		printf("invalid input.\n");
		return -1;
	}


	int type = atoi(argc[1]);
	Graph* graph;
	if (type == 1)
	{
		if (argv < 3)
		{
			printf("invalid input for type\n");
			return -1;
		}
		int size = atoi(argc[2]);
		graph = GenerateDepthStudyProblemInstance(size,size);
	}
	else if (type == 2)
	{
		if (argv < 7)
		{
			printf("invalid input for type. required: seed items constraints sizeLower sizeUpper\n");
			return -1;
		}
		int seed = atoi(argc[2]);
		int items = atoi(argc[3]);
		int constraints = atoi(argc[4]);
		float sizeLower = ((float)atoi(argc[5]))/100.0f;
		float sizeUpper = ((float)atoi(argc[6]))/100.0f;
		graph = GenerateRandomizedProblem(seed, items, constraints, sizeLower, sizeUpper);
	}
	else
	{
		printf("Unsupported type\n");
		return -1;
	}
	start = clock();
	BranchAndCut(graph);
	int ms = timePassed();
	printf("done - %i\n", ms);

	DeinitGraph(graph);

	return 0;
}

int* _res;
int _resHead;
int* _bestimate;
int _bestimateVal;
ListItem* _listStart;
Graph* _graph;

void Recurse(int step, int currentCost, int remainder, int totalRemainder, ListItem* nextInQueue);

void BranchAndCut(Graph* graph)
{
	int c = graph->Count;
	_res = malloc(sizeof(int)*c);
	_resHead = 0;
	_bestimate = malloc(sizeof(int)*c);
	_bestimateVal = c+1;
	_graph = graph;

	int totalWeight = 0;
	for (int i = 0; i < c;i++)
	{
		totalWeight += graph->Nodes[i].Weight;
	}
	printf("Lower bound: %i\n", (totalWeight+MaxWeight-1)/MaxWeight);

	ListItem* listStart = malloc(sizeof(ListItem));
	listStart->Previous = NULL;
	listStart->Next = NULL;
	listStart->NextAtDepth = NULL;
	listStart->Index = -1;
	listStart->Weight = INT_MAX;
	listStart->Depth = -1;
	_listStart = listStart;
	

	for (int j = 0; j < c; j++)
	{
		if (graph->EdgeCounters[j] == 0)
			InsertSorted(listStart, graph->Nodes[j].Weight, j, 0);
	}

	Recurse(0, 1, MaxWeight, totalWeight, NULL);
	#if 0
	printf("Best (%i):", _bestimateVal);
	for (int i = 0; i < c; i++)
		printf(" %i", _bestimate[i]);
	printf("\n");
	#endif

	free(_res);
	free(_bestimate);
	_graph = NULL;
}

void Recurse(int step, int currentCost, int remainder, int totalRemainder, ListItem* nextInQueue)
{
	if (currentCost >= _bestimateVal || _bestimateVal <= (currentCost+(totalRemainder-remainder+MaxWeight-1)/MaxWeight))
		return;
	int len = _graph->Count;
	if (step == len)
	{
		//printf("(%.2f,%i)", (float)timePassed()/1000.0f, currentCost);
		#if PRINT_COMBINATION >= 1
		printf("Combination _%i_ - %i :", currentCost, timePassed());
		#if PRINT_COMBINATION >= 2
		for (int i = 0; i < len; i++)
			printf(" %i", _res[i]);
		#endif
		printf("\n");
		#endif

		if (currentCost < _bestimateVal)
		{
			_bestimateVal = currentCost;
			for (int i = 0; i < len; i++)
				_bestimate[i] = _res[i];
		}
		return;
	}

	ListItem* it = nextInQueue;
	if (it == NULL)
		it = _listStart->Next;
	while (it != NULL)
	{
		if (it->Weight < remainder)
			break;
		it = it->Next;
	}
	if (it == NULL)
		it = _listStart->Next;

	while (it != NULL)
	{
		int i = it->Index;
	
		DetachItem(it);
		_res[_resHead++] = i;
		
		ListItem* firstAtDepth = NULL;
		ListItem* lastAtDepth = NULL;
		ListItem* biggestAtDepth = NULL;

		//Decrement edge counters and update potential nodes
		for (int j = 0; j < _graph->EdgeCounts[i]; j++)
		{
			int idx = _graph->Edges[i][j];
			_graph->EdgeCounters[idx]--;
			if (_graph->EdgeCounters[idx] == 0)
			{
				ListItem* item = InsertSorted(_listStart, _graph->Nodes[idx].Weight, idx, step+1);
				if (firstAtDepth == NULL)
				{
					firstAtDepth = item;
					lastAtDepth = item;
					biggestAtDepth = item;
				}
				else
				{
					lastAtDepth->NextAtDepth = item;
					lastAtDepth = item;
				}
				if (biggestAtDepth->Weight < item->Weight)
						biggestAtDepth = item;
			}
		}


		//Recurse
		int weight = it->Weight;
		ListItem* nxt = biggestAtDepth == NULL || it->Next == NULL || it->Next->Weight > biggestAtDepth->Weight ? it->Next : biggestAtDepth;
		if (remainder - weight < 0)
			Recurse(step + 1, currentCost + 1, MaxWeight - weight, totalRemainder - weight, nxt);
		else
			Recurse(step + 1, currentCost, remainder - weight, totalRemainder - weight, nxt);

		//Increment edge counters
		for (int j = 0; j < _graph->EdgeCounts[i]; j++)
		{
			_graph->EdgeCounters[_graph->Edges[i][j]]++;
		}

		//Remove added potential nodes
		ListItem* it2 = firstAtDepth;
		while (it2 != NULL)
		{
			ListItem* tmp = it2->NextAtDepth;
			RemoveItem(it2);
			it2 = tmp;
		}

		_resHead--;
		AttachItem(it);
		it = it->Next;
	}
}
